# -*- coding: utf-8 -*-
#
# Copyright (c) 2009-2014 Felix Krull <f_krull@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys
import traceback
from optparse import OptionParser

import gi

gi.require_version("Gst", "1.0")

from gi.repository import Gst  # noqa isort:skip

import rgain3.rgio  # noqa  isort:skip
from rgain3 import __version__  # noqa isort:skip


__all__ = [
    "Error",
    "init_gstreamer",
    "common_options",
]


class Error(Exception):
    def __init__(self, message, exc_info=None):
        super().__init__(message)
        # as long as instances are only constructed in exception handlers, this
        # should get us what we want
        self.exc_info = exc_info if exc_info else sys.exc_info()

    def __str__(self):
        if not self._output_full_exception():
            return super().__str__()
        else:
            return "".join(traceback.format_exception(*self.exc_info))

    def _output_full_exception(self):
        return self.exc_info[0] not in [
            IOError, rgain3.rgio.AudioFormatError, rgain3.GSTError,
        ]


def init_gstreamer():
    """Properly initialise GStreamer for the command-line interfaces.

    Specifically, GStreamer options are parsed and processed by GStreamer, but
    it is also kept from taking over the main help output (by pretending -h
    or --help wasn't passed, if necessary). --help-gst should be documented in
    the main help output as a switch to display GStreamer options."""
    # Strip any --help options from the command line.
    stripped_options = []
    for opt in ["-h", "--help"]:
        if opt in sys.argv:
            sys.argv.remove(opt)
            stripped_options.append(opt)
    # Then, pass any remaining options to GStreamer.
    sys.argv = Gst.init(sys.argv)
    # Finally, restore any help options so optparse can eat them.
    for opt in stripped_options:
        sys.argv.append(opt)


def common_options():
    opts = OptionParser(version="%%prog %s" % __version__)

    opts.add_option("-f", "--force", help="Recalculate Replay Gain even if the "
                    "file already contains gain information.", dest="force",
                    action="store_true")
    opts.add_option("-d", "--dry-run", help="Don't actually modify any files.",
                    dest="dry_run", action="store_true")
    opts.add_option("-r", "--reference-loudness", help="Set the reference "
                    "loudness to REF dB (default: %default dB)", metavar="REF",
                    dest="ref_level", action="store", type="int")
    opts.add_option("--mp3-format", help="Choose the Replay Gain data format "
                    "for MP3 files. The default setting should be compatible "
                    "with most decent software music players, so it is "
                    "generally not necessary to mess with this setting. Check "
                    "the README or man page for more information.",
                    dest="mp3_format", action="store", type="choice",
                    choices=rgain3.rgio.BaseFormatsMap.MP3_DISPLAY_FORMATS)
    # This option only exists to show up in the help output; if it's actually
    # specified, GStreamer should eat it.
    opts.add_option("--help-gst", help="Show GStreamer options.",
                    dest="help_gst", action="store_true")

    opts.set_defaults(
        force=False, dry_run=False, ref_level=89, mp3_format="default")

    return opts



# Copyright (c) 2009-2015 Felix Krull <f_krull@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import contextlib
import logging
import sys
from typing import Optional, Union

logger = logging.getLogger(__name__)


def _ensure_str(value: Union[str, bytes]) -> str:
    if isinstance(value, str):
        return value
    elif isinstance(value, bytes):
        return value.decode()
    raise TypeError


def _str_to_float(value: str) -> Optional[float]:
    try:
        return float(value)
    except ValueError:
        logger.info("Could not convert '%s' to float", value)
    return None


def parse_db(value: Union[str, bytes]) -> Optional[float]:
    # The db value from mutagen _should_ by of type str, but there are cases
    # where we get bytes, see https://github.com/chaudum/rgain3/issues/38
    value = _ensure_str(value.strip())
    if value.lower().endswith("db"):
        value = value[:-2].strip()
    return _str_to_float(value)


def parse_peak(value: str) -> Optional[float]:
    # The db value from mutagen _should_ by of type str, but there are cases
    # where we get bytes, see https://github.com/chaudum/rgain3/issues/38
    value = _ensure_str(value.strip())
    return _str_to_float(value)


def almost_equal(
    a: Optional[float], b: Optional[float], epsilon: float
) -> bool:
    if a is None and b is None:
        return True
    elif a is None or b is None:
        return False
    return abs(a - b) <= epsilon


def getfilesystemencoding():
    # get file system encoding, making sure never to return None
    return sys.getfilesystemencoding() or sys.getdefaultencoding()


@contextlib.contextmanager
def gobject_signals(obj, *signals):
    """Context manager to connect and disconnect GObject signals using a ``with``
    statement.
    """
    signal_ids = []
    try:
        for signal in signals:
            signal_ids.append(obj.connect(*signal))
        yield
    finally:
        for signal_id in signal_ids:
            obj.disconnect(signal_id)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2009-2014 Felix Krull <f_krull@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import abc
import os
import warnings

import mutagen
from mutagen.easyid3 import EasyID3
from mutagen.id3._util import ID3NoHeaderError

from rgain3 import GainData
from rgain3.util import almost_equal, parse_db, parse_peak


class AudioFormatError(Exception):
    def __init__(self, filename):
        super().__init__("Did not understand file: {}".format(filename))


# interface for ReplayGain reading/writing class
class BaseTagReaderWriter(abc.ABC):

    @abc.abstractmethod
    def read_gain(self, filename):
        raise NotImplementedError()

    @abc.abstractmethod
    def write_gain(self, filename, track_gain, album_gain):
        raise NotImplementedError()


# class to read and write ReplayGain data from/to simple tags. The default tags
# match the rg.org specification for Ogg (at least Vorbis), Flac and WavPack
# files
class SimpleTagReaderWriter(BaseTagReaderWriter):
    TRACK_GAIN_TAG = "replaygain_track_gain"
    TRACK_PEAK_TAG = "replaygain_track_peak"
    ALBUM_GAIN_TAG = "replaygain_album_gain"
    ALBUM_PEAK_TAG = "replaygain_album_peak"
    REF_LOUDNESS_TAGS = ["replaygain_reference_loudness"]

    # default behaviour; override in a subclass if necessary, e.g. for MP3
    def _get_tags_object(self, filename):
        return mutagen.File(filename)

    def read_gain(self, filename):
        tags = self._get_tags_object(filename)
        if tags is None:
            raise AudioFormatError(filename)

        track_gain = self._read_gain_data(tags, self.TRACK_GAIN_TAG,
                                          self.TRACK_PEAK_TAG)
        album_gain = self._read_gain_data(tags, self.ALBUM_GAIN_TAG,
                                          self.ALBUM_PEAK_TAG)
        ref_level = self._read_ref_loudness(tags)
        if ref_level is not None:
            if track_gain:
                track_gain.ref_level = ref_level
            if album_gain:
                album_gain.ref_level = ref_level
        return track_gain, album_gain

    def _read_gain_data(self, tags, gain_tag, peak_tag):
        if gain_tag in tags:
            gain = parse_db(tags[gain_tag][0])
            if gain is None:
                return None
            gaindata = GainData(gain)
            if peak_tag in tags:
                peak = parse_peak(tags[peak_tag][0])
                if peak is not None:
                    gaindata.peak = peak
        else:
            gaindata = None
        return gaindata

    def _read_ref_loudness(self, tags):
        for tag in self.REF_LOUDNESS_TAGS:
            if tag in tags:
                ref_level = parse_db(tags[tag][0])
                if ref_level is not None:
                    return ref_level
        return None

    def write_gain(self, filename, track_gain, album_gain):
        tags = self._get_tags_object(filename)
        if tags is None:
            raise AudioFormatError(filename)

        if track_gain:
            tags[self.TRACK_GAIN_TAG] = self._dump_gain(track_gain.gain)
            tags[self.TRACK_PEAK_TAG] = self._dump_peak(track_gain.peak)
            for tag in self.REF_LOUDNESS_TAGS:
                tags[tag] = self._dump_ref_level(track_gain.ref_level)

        if album_gain:
            tags[self.ALBUM_GAIN_TAG] = self._dump_gain(album_gain.gain)
            tags[self.ALBUM_PEAK_TAG] = self._dump_peak(album_gain.peak)

        tags.save()

    def _dump_gain(self, gain):
        return "%.8f dB" % gain

    def _dump_peak(self, peak):
        return "%.8f" % peak

    def _dump_ref_level(self, ref_level):
        return "%i dB" % ref_level


# MP4 support
class MP4TagReaderWriter(SimpleTagReaderWriter):
    FORMAT = "----:com.apple.iTunes:replaygain_%s_%s"
    TRACK_GAIN_TAG = FORMAT % ("track", "gain")
    TRACK_PEAK_TAG = FORMAT % ("track", "peak")
    ALBUM_GAIN_TAG = FORMAT % ("album", "gain")
    ALBUM_PEAK_TAG = FORMAT % ("album", "peak")
    REF_LOUDNESS_TAGS = []

    # Mutagen 1.22 has a bug (?) such that MP4 values cannot be unicode objects
    # so we encode everything to ASCII here
    # https://code.google.com/p/mutagen/issues/detail?id=164
    def _dump_gain(self, gain):
        return SimpleTagReaderWriter._dump_gain(self, gain).encode("ascii")

    def _dump_peak(self, peak):
        return SimpleTagReaderWriter._dump_peak(self, peak).encode("ascii")

    def _dump_ref_level(self, ref_level):
        return SimpleTagReaderWriter._dump_ref_level(
            self, ref_level).encode("ascii")


# MP3 support base class
class MP3TagReaderWriter(SimpleTagReaderWriter):
    _EXTRA_TXXX_TAGS = [
        "replaygain_track_gain",
        "replaygain_track_peak",
        "replaygain_album_gain",
        "replaygain_album_peak",
        "replaygain_reference_loudness",
        "QuodLibet::replaygain_reference_loudness",
    ]

    class _ReplaygainEasyID3(EasyID3):
        pass

    for key in _EXTRA_TXXX_TAGS:
        _ReplaygainEasyID3.RegisterTXXXKey("TXXX:%s" % key, key)

    def _get_tags_object(self, filename):
        try:
            return self._ReplaygainEasyID3(filename)
        except ID3NoHeaderError:
            tags = self._ReplaygainEasyID3()
            tags.filename = filename
            return tags


# ID3v2 support for TXXX:replaygain_* frames as specified in
# http://wiki.hydrogenaudio.org/index.php?title=ReplayGain_specification#ID3v2
# and as implemented by at least foobar2000.
class MP3rgorgTagReaderWriter(MP3TagReaderWriter):
    TRACK_GAIN_TAG = "TXXX:replaygain_track_gain"
    TRACK_PEAK_TAG = "TXXX:replaygain_track_peak"
    ALBUM_GAIN_TAG = "TXXX:replaygain_album_gain"
    ALBUM_PEAK_TAG = "TXXX:replaygain_album_peak"
    REF_LOUDNESS_TAGS = ["TXXX:replaygain_reference_loudness"]


# clamp RVA2 values to certain limits so that they do not overflow
RVA2_GAIN_MIN = -64
RVA2_GAIN_MAX = float(64 * 512 - 1) / 512.0
RVA2_PEAK_MIN = 0
RVA2_PEAK_MAX = float(2 ** 16 - 1) / float(2 ** 15)


def clamp(v, min, max):
    clamped = False
    if v < min:
        v = min
        clamped = True
    if v > max:
        v = max
        clamped = True
    return v, clamped


def clamp_rva2_gain(v):
    v, clamped = clamp(v, RVA2_GAIN_MIN, RVA2_GAIN_MAX)
    if clamped:
        warnings.warn("gain value was out of bounds for RVA2 frame and was "
                      "clamped to %.2f" % v)
    return v


# I'm not sure if this situation could even reasonably happen, but
# can't hurt to check, right? Right!?
def clamp_rva2_peak(v):
    v, clamped = clamp(v, RVA2_PEAK_MIN, RVA2_PEAK_MAX)
    if clamped:
        warnings.warn("peak value was out of bounds for RVA2 frame and was "
                      "clamped to %.5f" % v)
    return v


def clamp_gain_data(gain_data):
    if gain_data is None:
        return None
    else:
        return GainData(clamp_rva2_gain(gain_data.gain),
                        clamp_rva2_peak(gain_data.peak),
                        gain_data.ref_level)


# ID3v2 support for legacy RVA2-frames-based format according to
# http://wiki.hydrogenaudio.org/index.php?title=ReplayGain_specification#ID3v2
class MP3RVA2TagReaderWriter(MP3TagReaderWriter):
    # EasyID3 maps these to RVA2 by default
    TRACK_GAIN_TAG = "replaygain_track_gain"
    TRACK_PEAK_TAG = "replaygain_track_peak"
    ALBUM_GAIN_TAG = "replaygain_album_gain"
    ALBUM_PEAK_TAG = "replaygain_album_peak"

    # since there's no proper reference loudness tag for the legacy format, we
    # use reasonably common TXXX tags; these were registered in the superclass
    REF_LOUDNESS_TAGS = [
        "TXXX:replaygain_reference_loudness",
        "TXXX:QuodLibet::replaygain_reference_loudness",
    ]

    def _dump_gain(self, gain):
        return SimpleTagReaderWriter._dump_gain(self, clamp_rva2_gain(gain))

    def _dump_peak(self, peak):
        return SimpleTagReaderWriter._dump_peak(self, clamp_rva2_peak(peak))


# Special compatible MP3 support that
#  - reads both rg.org and legacy gain, compares them, returns them if they
#    match, else returns no gain data
#  - writes both rg.org and legacy gain
class MP3DefaultTagReaderWriter(BaseTagReaderWriter):
    def __init__(self, rgorg_readerwriter, rva2_readerwriter):
        self.rgorg = rgorg_readerwriter
        self.rva2 = rva2_readerwriter

    def read_gain(self, filename):
        rgorg_track_gain, rgorg_album_gain = self.rgorg.read_gain(filename)
        rva2_track_gain, rva2_album_gain = self.rva2.read_gain(filename)
        # We want to ensure we have all bits of data so if we only have one
        # format, we say we have none to enforce recalculation.
        if rgorg_track_gain is None or rva2_track_gain is None:
            # ensure that track gain exists for both
            return (None, None)
        if (not gaindata_almost_equal(rgorg_track_gain, rva2_track_gain) or
                not gaindata_almost_equal(rgorg_album_gain, rva2_album_gain)):
            # The different formats are not similar enough.
            return (None, None)
        else:
            # the formats seem to match, we obviously use the non-clamped one
            return (rgorg_track_gain, rgorg_album_gain)

    def write_gain(self, filename, track_gain, album_gain):
        self.rgorg.write_gain(filename, track_gain, album_gain)
        self.rva2.write_gain(filename, track_gain, album_gain)


GAIN_EPSILON = 0.1
PEAK_EPSILON = 0.001
REF_LEVEL_EPSILON = 0.1


# For these three functions, b is always the legacy values, i.e. the
# potentially clamped ones.
def gain_almost_equal(a, b):
    return (almost_equal(a, b, GAIN_EPSILON) or
            almost_equal(clamp_rva2_gain(a), b, GAIN_EPSILON))


def peak_almost_equal(a, b):
    return (almost_equal(a, b, PEAK_EPSILON) or
            almost_equal(clamp_rva2_peak(a), b, PEAK_EPSILON))


def gaindata_almost_equal(a, b):
    # Ensure neither element is None.
    if a is None:
        return b is None
    if b is None:
        return a is None
    if a == b:
        return True

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        return (gain_almost_equal(a.gain, b.gain) and
                peak_almost_equal(a.peak, b.peak) and
                almost_equal(a.ref_level, b.ref_level, REF_LEVEL_EPSILON))


# code to pull everything together
class UnknownFiletype(Exception):
    pass


class BaseFormatsMap(object):
    _simplereaderwriter = SimpleTagReaderWriter()
    _mp4readerwriter = MP4TagReaderWriter()
    _mp3_rgorg_readerwriter = MP3rgorgTagReaderWriter()
    _mp3_rva2_readerwriter = MP3RVA2TagReaderWriter()
    _mp3_default_readerwriter = MP3DefaultTagReaderWriter(
        _mp3_rgorg_readerwriter,
        _mp3_rva2_readerwriter)

    BASE_MAP = {
        ".ogg": _simplereaderwriter,
        ".oga": _simplereaderwriter,
        ".flac": _simplereaderwriter,
        ".wv": _simplereaderwriter,
        ".m4a": _mp4readerwriter,
        ".mp4": _mp4readerwriter,
    }

    MP3_FORMATS = {
        None: _mp3_default_readerwriter,
        "default": _mp3_default_readerwriter,
        "replaygain.org": _mp3_rgorg_readerwriter,
        "fb2k": _mp3_rgorg_readerwriter,
        "legacy": _mp3_rva2_readerwriter,
        "ql": _mp3_rva2_readerwriter,
    }

    MP3_DISPLAY_FORMATS = ["default", "replaygain.org", "legacy", "ql", "fb2k"]
    MP3_DEFAULT_FORMAT = "default"

    def __init__(self, mp3_format=None, more_mappings=None):
        # yeah, you need to choose
        self.more_mappings = more_mappings if more_mappings else {}
        if mp3_format in self.MP3_FORMATS:
            self.more_mappings[".mp3"] = self.MP3_FORMATS[mp3_format]
        else:
            raise ValueError("invalid MP3 format %r" % mp3_format)

    def is_supported_format(self, ext):
        ext_lower = ext.lower()
        return ext_lower in self.BASE_MAP or ext_lower in self.more_mappings

    def read_gain(self, filename):
        ext = os.path.splitext(filename)[1].lower()
        if ext in self.more_mappings:
            accessor = self.more_mappings[ext]
        elif ext in self.BASE_MAP:
            accessor = self.BASE_MAP[ext]
        else:
            raise UnknownFiletype(ext)

        return accessor.read_gain(filename)

    def write_gain(self, filename, trackgain, albumgain):
        ext = os.path.splitext(filename)[1].lower()
        if ext in self.more_mappings:
            accessor = self.more_mappings[ext]
        elif ext in self.BASE_MAP:
            accessor = self.BASE_MAP[ext]
        else:
            raise UnknownFiletype(ext)

        accessor.write_gain(filename, trackgain, albumgain)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2009-2014 Felix Krull <f_krull@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import contextlib
import io
import multiprocessing
import os.path
import pickle
import sys
from hashlib import md5

import mutagen

from rgain3 import albumid, rgio
from rgain3.script import Error, common_options, init_gstreamer
from rgain3.script.replaygain import do_gain

CURRENT_CACHE_VERSION = 1


# all of collectiongain
def relpath(path, base):
    size = len(base)
    if os.path.basename(base):
        # this means ``base`` does not end with a separator
        size += 1
    return path[size:]


def cache_entry_valid(filepath, record):
    return (
        isinstance(filepath, str) and
        hasattr(record, "__getitem__") and
        hasattr(record, "__len__") and
        len(record) == 3 and
        (isinstance(record[0], str) or record[0] is None) and
        (isinstance(record[1], int) or isinstance(record[1], float)) and
        isinstance(record[2], bool))


def read_cache(cache_file):
    if os.path.isfile(cache_file):
        try:
            with open(cache_file, "rb") as f:
                cache = pickle.load(f)
                if not isinstance(cache, tuple) or len(cache) != 2:
                    print("Invalid cache, ignoring it")
                    return {}
                cache_version = cache[0]
                if cache_version != CURRENT_CACHE_VERSION:
                    print("Old cache format, ignoring it")
                    return {}
                files = cache[1]
                if not isinstance(files, dict):
                    print("Invalid cache, ignoring it")
                    return {}
                to_remove = set()
                for filepath, record in files.items():
                    if not cache_entry_valid(filepath, record):
                        to_remove.add(filepath)
                for filepath in to_remove:
                    # remove fishy entries
                    del files[filepath]
                return files
        except Exception as exc:
            print(
                "Error while reading the cache, continuing without it - %s" %
                (exc,))
    return {}


def write_cache(cache_file, files):
    cache_dir = os.path.dirname(cache_file)
    try:
        if not os.path.isdir(cache_dir):
            os.makedirs(cache_dir, 0o755)
        with open(cache_file, "wb") as f:
            pickle.dump((CURRENT_CACHE_VERSION, files), f, 2)
    except Exception as exc:
        print("Error while writing the cache - %s" % exc)


def collect_files(music_dir, files, visited_cache, is_supported_format):
    i = 0
    for dirpath, dirnames, filenames in os.walk(music_dir):
        for filename in filenames:
            filepath = relpath(os.path.join(dirpath, filename), music_dir)
            properpath = os.path.join(dirpath, filename)
            mtime = os.path.getmtime(properpath)

            # check the cache
            if filepath in visited_cache:
                visited_cache[filepath] = True
                record = files[filepath]
                if mtime <= record[1]:
                    # the file's still ok
                    continue

            ext = os.path.splitext(filename)[1]
            if is_supported_format(ext):
                i += 1
                print("  [%i] %s |" % (i, filepath), end='')
                try:
                    tags = mutagen.File(os.path.join(music_dir, filepath))
                    if tags is None:
                        raise Exception()
                    album_id = albumid.get_album_id(tags)
                    print(album_id or "<single track>")
                    # fields here: album_id, mtime, already_processed
                    files[filepath] = (album_id, mtime, False)
                except Exception:
                    # TODO: Maybe optionally abort here?
                    print("IGNORED: unreadable file or unsupported format")


def transform_cache(files):
    # transform ``files`` into lists of things to process
    albums = {}
    single_tracks = []
    for filepath, (album_id, mtime, processed) in files.items():
        if album_id is not None:
            albums.setdefault(album_id, []).append(filepath)
        else:
            single_tracks.append(filepath)

    # purge anything that's marked as processed
    for album_id, album_files in list(albums.items()):
        keep = False
        for filepath in album_files:
            if not files[filepath][2]:
                keep = True
                break
        if not keep:
            del albums[album_id]

    for filepath in single_tracks[:]:
        if files[filepath][2]:
            single_tracks.remove(filepath)

    return albums, single_tracks


def update_cache(files, music_dir, tracks, album_id):
    for filepath in tracks:
        mtime = os.path.getmtime(os.path.join(music_dir, filepath))
        files[filepath] = (album_id, mtime, True)


@contextlib.contextmanager
def stdstreams(stdout, stderr):
    old_stdout = sys.stdout
    old_stderr = sys.stderr
    sys.stdout = stdout
    sys.stderr = stderr
    try:
        yield
    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr


def do_gain_async(queue, job_key, files, ref_level, force, dry_run, album,
                  mp3_format):
    output = io.StringIO()
    try:
        with stdstreams(output, output):
            if album:
                print("%s:" % job_key[1], end='')
            do_gain(files, ref_level, force, dry_run, album, mp3_format)
            print("")
    except BaseException as exc:
        # We can't reliably serialise and pass the exception information to the
        # driver process so we stringify it here.
        # And yes, we want to catch KeyboardInterrupt et al.
        queue.put((job_key, output.getvalue(), str(exc)))
    else:
        queue.put((job_key, output.getvalue(), None))


def do_gain_all(music_dir, albums, single_tracks, files, ref_level=89,
                force=False, dry_run=False, mp3_format=None, jobs=0,
                stop_on_error=False):
    pool = multiprocessing.Pool(None if jobs == 0 else jobs)
    manager = multiprocessing.Manager()
    queue = manager.Queue()
    num_jobs = 0

    print("Dispatching jobs ...")
    if single_tracks:
        pool.apply_async(
            do_gain_async, [
                queue, (single_tracks, None),
                [os.path.join(music_dir, path) for path in single_tracks],
                ref_level, force, dry_run, False, mp3_format])
        num_jobs += 1

    for album_id, album_files in albums.items():
        pool.apply_async(
            do_gain_async, [
                queue, (album_files, album_id),
                [os.path.join(music_dir, path) for path in album_files],
                ref_level, force, dry_run, True, mp3_format])
        num_jobs += 1
    pool.close()

    print("Now waiting for results ...")
    failed_jobs = []
    try:
        all_jobs = num_jobs
        successful = 0
        while num_jobs > 0:
            job_key, output, exc = queue.get()
            num_jobs -= 1
            if exc:
                failed_jobs.append((job_key, output, exc))
            else:
                successful += 1
                print(output.strip())
                print(
                    "Successfully finished %s of %s." % (successful, all_jobs))
                print("")
            # Update cache.
            if not dry_run:
                tracks, album_id = job_key
                update_cache(files, music_dir, tracks, album_id)
    finally:
        try:
            pool.terminate()
        except Exception:
            # terminate ends rather horribly so we just silence it.
            pass
        if len(failed_jobs) > 0:
            print("Unfortunately, there were some errors:")
            for key, output, exc in failed_jobs:
                print(output.strip())
                print(exc, file=sys.stderr)
                print("")
        print("%s successful, %s failed." % (successful, len(failed_jobs)))


def do_collectiongain(music_dir, ref_level=89, force=False, dry_run=False,
                      mp3_format=None, ignore_cache=False, jobs=0):
    music_abspath = os.path.abspath(music_dir)
    musicpath_hash = md5(music_abspath.encode("utf-8")).hexdigest()
    cache_file = os.path.join(os.path.expanduser("~"), ".cache",
                              "collectiongain-cache.%s" % musicpath_hash)

    # load the cache, if desired
    if not ignore_cache:
        files = read_cache(cache_file)
    else:
        files = {}

    print("Collecting files ...")
    # whenever this part is stopped (KeyboardInterrupt/other exception), the
    # cache is written to disk so all progress persists
    try:
        visited_cache = dict.fromkeys(iter(files.keys()), False)
        collect_files(music_dir, files, visited_cache,
                      rgio.BaseFormatsMap(mp3_format).is_supported_format)
        # clean cache
        for filepath, visited in list(visited_cache.items()):
            if not visited:
                del visited_cache[filepath]
                del files[filepath]
        # hopefully gets rid of at least one huge data structure
        del visited_cache

        albums, single_tracks = transform_cache(files)

        # gain everything that has survived the cleansing
        do_gain_all(
            music_dir, albums, single_tracks, files, ref_level, force, dry_run,
            mp3_format, jobs)
    finally:
        write_cache(cache_file, files)

    print("All finished.")


def collectiongain_options():
    opts = common_options()

    opts.add_option("--ignore-cache", help="Do not use the file cache at all.",
                    dest="ignore_cache", action="store_true")
    opts.add_option("--regain", help="Fully reprocess everything. Same as "
                    "'--force --ignore-cache'.",
                    dest="regain", action="store_true")
    opts.add_option("-j", "--jobs", help="Specifies the number of jobs to run "
                    "simultaneously. Must be >= 1. By default, this is set to "
                    "the number of CPU cores in the system to provide best "
                    "performance.", dest="jobs", action="store", type="int")

    opts.set_defaults(ignore_cache=False, jobs=None)

    opts.set_usage("%prog [options] MUSIC_DIR")
    opts.set_description("Calculate Replay Gain for a large set of audio files "
                         "without asking many questions. This program "
                         "calculates an album ID for any audo file in "
                         "MUSIC_DIR. Then, album gain will be applied to all "
                         "files with the same album ID. The album ID is "
                         "created from file tags as follows: If an 'album' tag "
                         "is present, it is joined with the contents of an "
                         "'albumartist' tag, or, if that isn't set, the "
                         "contents of the 'artist' tag, or nothing if there is "
                         "no 'artist' tag as well. On the other hand, if no "
                         "'album' tag is present, the file is assumed to be a "
                         "single track without album; in that case, no album "
                         "gain will be calculated for that file.")

    return opts


def collectiongain():
    init_gstreamer()
    optparser = collectiongain_options()
    opts, args = optparser.parse_args()
    if len(args) != 1:
        optparser.error("pass one directory path")
    if opts.jobs is not None and opts.jobs < 1:
        optparser.error("jobs must be at least 1")
    if opts.regain:
        opts.force = opts.ignore_cache = True

    try:
        do_collectiongain(args[0], opts.ref_level, opts.force, opts.dry_run,
                          opts.mp3_format, opts.ignore_cache, opts.jobs)
    except Error as exc:
        print("")
        print(str(exc), file=sys.stderr)
        sys.exit(1)
    except KeyboardInterrupt:
        print("Interrupted.")


if __name__ == "__main__":
    collectiongain()
